"""This module contains the Model class.

Authors: Thomas Oberbichler, Armin Geiser
"""

from collections import OrderedDict
from copy import deepcopy
from enum import Enum

import numpy as np
import numpy.linalg as la

from scipy.linalg import eig

from nfem.dof import Dof
from nfem.node import Node
from nfem.truss import Truss
from nfem.spring import Spring

from nfem.assembler import Assembler
from nfem.newton_raphson import newton_raphson_solve

from nfem.path_following_method import ArcLengthControl, DisplacementControl, LoadControl


class Model:
    """A Model contains all the objects that build the finite element model.
        Nodes, elements, loads, dirichlet conditions...

    Attributes
    ----------
    name : str
        Name of the model.
    nodes : dict
        Dictionary that stores node_id : node object
    elements : str
        Dictionary that stores element_id : element object
    lam : float
        load factor
    previous_model : Model
        Previous state of this model
    """

    def __init__(self, name=None):
        """Create a new model.

        Parameters
        ----------
        name : str
            Name of the model.
        """

        self.name = name
        self.status = ModelStatus.initial
        self.nodes = KeyCollection()
        self.elements = KeyCollection()
        self.load_factor = 0.0
        self._previous_model = None
        self.det_k = None
        self.first_eigenvalue = None
        self.first_eigenvector_model = None

    # === modeling

    def add_node(self, id, x, y, z=None, support='', fx=0.0, fy=0.0, fz=0.0):
        """Add a three dimensional node to the model.

        Parameters
        ----------
        id : str
            Unique ID of the node.
        x : float
            X coordinate.
        y : float
            Y coordinate.
        z : float
            Z coordinate.
        support : str
            Directions in which the displacements are fixed.
        fx : float
            External force in x direction.
        fy : float
            External force in y direction.
        fz : float
            External force in z direction.

        Returns
        ----------
        Node
            The new Node

        Examples
        --------
        Add a node with ID `B`:

        >>> model.add_node(id='B', x=5, y=2, z=0)
        """
        if not isinstance(id, str):
            raise TypeError('The node id is not a text string')

        if id in self.nodes:
            raise KeyError('The model already contains a node with id {}'.format(id))

        node = Node(id, x, y, 0 if z is None else z)

        self.nodes._add(node)

        if 'x' in support:
            node.dof('u').is_active = False
        if 'y' in support:
            node.dof('v').is_active = False
        if z is None or 'z' in support:
            node.dof('w').is_active = False

        node.dof('u').external_force = fx
        node.dof('v').external_force = fy
        node.dof('w').external_force = fz

    def add_truss(self, id, node_a, node_b, youngs_modulus, area, prestress=0):
        """Add a three dimensional truss element to the model.

        Parameters
        ----------
        id : str
            Unique ID of the element.
        node_a : str
            ID of the first node.
        node_b : str
            ID of the second node.
        youngs_modulus : float
            Youngs modulus of the material for the truss.
        area : float
            Area of the cross section for the truss.

        Returns
        ----------
        Truss
            The new Truss

        Examples
        --------
        Add a truss element from node `A` to node `B`:

        >>> model.add_truss(node_a='A', node_a='B', youngs_modulus=20, area=1)
        """
        if not isinstance(id, str):
            raise TypeError('The element id is not a text string')

        if not isinstance(node_a, str):
            raise TypeError('The id of node_a is not a text string')

        if not isinstance(node_b, str):
            raise TypeError('The id of node_b is not a text string')

        if id in self.elements:
            raise KeyError('The model already contains an element with id {}'.format(id))

        if node_a not in self.nodes:
            raise KeyError('The model does not contain a node with id {}'.format(node_a))

        if node_b not in self.nodes:
            raise KeyError('The model does not contain a node with id {}'.format(node_b))

        element = Truss(id, self.nodes[node_a], self.nodes[node_b], youngs_modulus, area, prestress)

        self.elements._add(element)

    def add_spring(self, id, node, kx=0, ky=0, kz=0):
        if not isinstance(id, str):
            raise TypeError('The element id is not a text string')

        if not isinstance(node, str):
            raise TypeError('The id of node is not a text string')

        if id in self.elements:
            raise KeyError('The model already contains an element with id {}'.format(id))

        if node not in self.nodes:
            raise KeyError('The model does not contain a node with id {}'.format(node))

        element = Spring(id, self.nodes[node], kx, ky, kz)

        self.elements._add(element)

    # === degree of freedoms

    def __getitem__(self, key):
        if isinstance(key, Dof):
            node_key, dof_type = key.id
        else:
            node_key, dof_type = key
        return self.nodes[node_key].dof(dof_type)

    @property
    def dofs(self):
        return Assembler(self).dofs

    # === increment

    def get_dof_increment(self, dof):
        """Get the increment of the dof during the last solution step

        Parameters
        ----------
        dof : tuple(node_id, dof_type)
            Dof that is asked

        Returns
        -------
        delta : float
            Increment of the dof during the last step
        """
        if self.get_previous_model() is None:
            return 0.0

        previous_model = self.get_previous_model()

        current_value = self[dof].delta
        previous_value = previous_model[dof].delta

        return current_value - previous_value

    def get_lam_increment(self):
        """Get the increment of lambda during the last solution step

        Returns
        -------
        delta : float
            Increment of lambda during the last step
        """
        if self.get_previous_model() is None:
            return 0.0

        current_value = self.load_factor
        previous_value = self.get_previous_model().load_factor

        return current_value - previous_value

    def get_increment_vector(self, assembler=None):
        """Get the increment that resulted in the current position
        """

        if assembler is None:
            assembler = Assembler(self)

        dof_count = assembler.dof_count

        increment = np.zeros(dof_count + 1)

        if self.get_previous_model() is None:
            print('WARNING: Increment is zero because no previous model exists!')
            return increment

        for index, dof in enumerate(assembler.dofs):
            increment[index] = self.get_dof_increment(dof)

        increment[-1] = self.get_lam_increment()

        return increment

    def get_increment_norm(self, assembler=None):
        increment = self.get_increment_vector(assembler)
        return la.norm(increment)

    # === model history

    def get_previous_model(self, skip_iterations=True):
        """Get the previous model of the current model.

        Parameters
        ----------
        skip_iterations : bool
            Flag if iteration or predicted previous models should be skipped

        Returns
        -------
        model : Model
            The previous model object
        """
        if not skip_iterations:
            return self._previous_model

        # find the most previous model that is not an iteration or prediction
        previous_model = self._previous_model

        while previous_model is not None and previous_model.status in [ModelStatus.duplicate,
                                                                       ModelStatus.prediction, ModelStatus.iteration]:
            previous_model = previous_model._previous_model

        return previous_model

    def get_initial_model(self):
        """Gets the initial model of this model.

        Returns
        ----------
        model : Model
            initial model
        """
        current_model = self

        while current_model.get_previous_model() is not None:
            current_model = current_model.get_previous_model()

        return current_model

    def get_model_history(self, skip_iterations=True):
        """Gets a list of all previous models of this model.

        Parameters
        ----------
        skip_iterations : bool
            Flag that decides if non converged previous models should be considered

        Returns
        ----------
        history : list
            List of previous models
        """

        history = [self]

        current_model = self

        while current_model.get_previous_model(skip_iterations) is not None:
            current_model = current_model.get_previous_model(skip_iterations)

            history = [current_model] + history

        return history

    def get_duplicate(self, name=None, branch=False):
        r"""Get a duplicate of the model.

        Parameters
        ----------
        name : str, optional
            Name of the new model.
        branch : bool, optional
            If `branch` is `False` the duplicate will be a successor of the current model::

                previous ----> current ----> duplicate

            If `branch` is `True` the duplicate will be a successor of the previous model::

                previous ----> current
                          \
                           \-> duplicate

        Returns
        -------
        model : Model
            Duplicate of the current model.
        """

        temp_previous_model = self._previous_model
        self._previous_model = None

        duplicate = deepcopy(self)

        self._previous_model = temp_previous_model

        if branch:
            duplicate._previous_model = self._previous_model
        else:
            duplicate._previous_model = self
            duplicate.status = ModelStatus.duplicate

        if name is not None:
            duplicate.name = name

        # make sure the duplicated model is in a clean state
        duplicate.det_k = None
        duplicate.first_eigenvalue = None
        duplicate.first_eigenvector_model = None

        return duplicate

    def new_timestep(self, name=None):
        temp_previous_model = self._previous_model
        self._previous_model = None

        duplicate = deepcopy(self)

        self._previous_model = temp_previous_model

        duplicate._previous_model = self

        if name is not None:
            duplicate.name = name

        duplicate.det_k = None
        duplicate.first_eigenvalue = None
        duplicate.first_eigenvector_model = None

        return duplicate

    # === solving

    def perform_linear_solution_step(self):
        """Performs a linear solution step on the model.
            It uses the current load factor.
            The results are stored at the dofs and used to update the current
            coordinates of the nodes.
        """

        print("\n=================================")
        print("Start linear solution step...")
        print("lambda : {}".format(self.load_factor))

        assembler = Assembler(self)

        dof_count = assembler.dof_count

        u = np.zeros(dof_count)

        for dof in assembler.dofs:
            index = assembler.index_of_dof(dof)
            u[index] = self[dof].delta

        k = np.zeros((dof_count, dof_count))
        f = np.zeros(dof_count)

        for i, dof in enumerate(assembler.dofs):
            f[i] += self[dof].external_force

        assembler.assemble_matrix(k, lambda element: element.calculate_elastic_stiffness_matrix())

        f *= self.load_factor

        try:
            u = la.solve(k, f)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        for index, dof in enumerate(assembler.dofs):
            self[dof].delta = u[index]

        self.status = ModelStatus.equilibrium

    def perform_non_linear_solution_step(self, strategy, tolerance=1e-5, max_iterations=100, **options):
        """Performs a non linear solution step on the model.
            The path following strategy is chose according to the parameter.
            A newton raphson algorithm is used to iteratively solve the nonlinear
            equation system r(u,lam) = 0
            The results are stored at the dofs and used to update the current
            coordinates of the nodes.

        Parameters
        ----------
        strategy : string
            Path following strategy. Available options:
            - load-control
            - displacement-control
            - arc-length-control
        max_iterations: int
            Maximum number of iteration for the newton raphson
        tolerance : float
            Tolerance for the newton raphson
        **options: kwargs (key word arguments)
            Additional options e.g.
            - dof=('B','v'): for displacement-control
            - solve_det_k=True: for solving the determinant of k at convergence
            - solve_attendant_eigenvalue=True: for solving the attendant eigenvalue problem at convergence
        """

        print("\n=================================")
        print("Start non linear solution step...")

        if strategy == 'load-control':
            constraint = LoadControl(self)
        elif strategy == 'displacement-control':
            constraint = DisplacementControl(self, options['dof'])
        elif strategy == 'arc-length-control':
            constraint = ArcLengthControl(self)
        else:
            raise ValueError('Invalid path following strategy:' + strategy)

        # initialize working matrices and functions for newton raphson
        assembler = Assembler(self)
        dof_count = assembler.dof_count

        def calculate_system(x):
            """Callback function for the newton raphson method that calculates the
                system for a given state x

            Parameters
            ----------
            x : ndarray
                Current state of dofs and lambda (unknowns of the non linear system)

            Returns
            ----------
            lhs : ndarray
                Left hand side matrix of size (n_free_dofs+1,n_free_dofs+1).
                Containts the derivatives of the residuum and the constraint.
            rhs : ndarray
                Right hand side vector of size (n_free_dofs+1).
                Contains the values of the residuum of the structure and the constraint.
            """

            # create a duplicate of the current state before updating and insert it in the history
            duplicate = self.get_duplicate()
            duplicate._previous_model = self._previous_model
            self._previous_model = duplicate
            duplicate.status = self.status

            # update status flag
            self.status = ModelStatus.iteration

            # update actual coordinates
            for index, dof in enumerate(assembler.dofs):
                self[dof].delta = x[index]

            # update lambda
            self.load_factor = x[-1]

            # initialize with zeros
            k = np.zeros((dof_count, dof_count))
            external_f = np.zeros(dof_count)
            internal_f = np.zeros(dof_count)

            # assemble stiffness
            assembler.assemble_matrix(k, lambda element: element.calculate_stiffness_matrix())

            # assemble force

            for i, dof in enumerate(assembler.dofs):
                external_f[i] += self[dof].external_force

            assembler.assemble_vector(internal_f, lambda element: element.calculate_internal_forces())

            # assemble left and right hand side for newton raphson
            lhs = np.zeros((dof_count + 1, dof_count + 1))
            rhs = np.zeros(dof_count + 1)

            # mechanical system
            lhs[:dof_count, :dof_count] = k
            lhs[:dof_count, -1] = -external_f
            rhs[:dof_count] = internal_f - self.load_factor * external_f

            # assemble contribution from constraint
            constraint.calculate_derivatives(self, lhs[-1, :])
            rhs[-1] = constraint.calculate_constraint(self)

            return lhs, rhs

        # prediction as vector for newton raphson
        x = np.zeros(dof_count + 1)
        for index, dof in enumerate(assembler.dofs):
            x[index] = self[dof].delta

        x[-1] = self.load_factor

        # solve newton raphson
        x, n_iter = newton_raphson_solve(calculate_system, x, max_iterations, tolerance)

        print("Solution found after {} iteration steps.".format(n_iter))

        self.status = ModelStatus.equilibrium

        if 'solve_det_k' in options and not options['solve_det_k']:
            pass
        else:
            self.solve_det_k(assembler=assembler)

        if 'solve_attendant_eigenvalue' in options:
            if options['solve_attendant_eigenvalue']:
                self.solve_eigenvalues(assembler=assembler)

    def solve_det_k(self, k=None, assembler=None):
        """Solves the determinant of k

        Parameters
        ----------
        k : numpy.ndarray (optional)
            stiffness matrix can be directly passed.
        assembler : Object (optional)
            assembler can be passed to speed up if k is not given
        """
        if k is None:
            if assembler is None:
                assembler = Assembler(self)
            dof_count = assembler.dof_count
            k = np.zeros((dof_count, dof_count))
            assembler.assemble_matrix(k, lambda element: element.calculate_stiffness_matrix())
        self.det_k = la.det(k)
        print("Det(K): {}".format(self.det_k))

    def solve_linear_eigenvalues(self, assembler=None):
        """Solves the linearized eigenvalue problem
           [ k_e + eigvals * k_g(linear strain) ] * eigvecs = 0
           Stores the first positive eigenvalue and vector

        Parameters
        ----------
        assembler : Object (optional)
            assembler can be passed to speed up
        """
        if assembler is None:
            assembler = Assembler(self)

        dof_count = assembler.dof_count

        # assemble matrices
        k_e = np.zeros((dof_count, dof_count))
        k_g = np.zeros((dof_count, dof_count))
        print("=================================")
        print('Linearized prebuckling (LPB) analysis ...')
        assembler.assemble_matrix(k_e, lambda element: element.calculate_elastic_stiffness_matrix())
        assembler.assemble_matrix(k_g, lambda element: element.calculate_geometric_stiffness_matrix(linear=True))

        # solve eigenvalue problem
        eigvals, eigvecs = eig(k_e, -k_g)

        # extract real parts of eigenvalues
        eigvals = np.array([x.real for x in eigvals])

        # sort eigenvalues and vectors
        idx = eigvals.argsort()
        eigvals = eigvals[idx]
        eigvecs = eigvecs[:, idx]

        # remove negative eigenvalues
        for i, eigenvalue in enumerate(eigvals):
            if eigenvalue > 0.0:
                break
            else:
                i += 1

        if i == len(eigvals):
            print('System has no positive eigenvalues!')
            return

        eigvals = eigvals[i:]
        eigvecs = eigvecs[:, i:]

        print('First linear eigenvalue: {}'.format(eigvals[0]))
        print('First linear eigenvalue * lambda: {}'.format(eigvals[0] * self.load_factor))  # this is printed in TRUSS
        if len(eigvecs[0]) < 10:
            print('First linear eigenvector: {}'.format(eigvecs[0]))

        self.first_eigenvalue = eigvals[0]

        # store eigenvector as model
        model = self.get_duplicate()
        model._previous_model = self
        model.status = ModelStatus.eigenvector
        model.det_k = None
        model.first_eigenvalue = None
        model.first_eigenvector_model = None
        model.load_factor = None

        for index, dof in enumerate(assembler.dofs):
            model[dof].delta = eigvecs[index][0]

        self.first_eigenvector_model = model

    def solve_eigenvalues(self, assembler=None):
        """Solves the eigenvalue problem
           [ k_m + eigvals * k_g ] * eigvecs = 0
           Stores the closest (most critical) eigenvalue and vector

        Parameters
        ----------
        assembler : Object (optional)
            assembler can be passed to speed up
        """
        if assembler is None:
            assembler = Assembler(self)

        dof_count = assembler.dof_count

        # assemble matrices
        k_m = np.zeros((dof_count, dof_count))
        k_g = np.zeros((dof_count, dof_count))
        print("=================================")
        print('Attendant eigenvalue analysis ...')
        assembler.assemble_matrix(k_m, lambda element: element.calculate_material_stiffness_matrix())
        assembler.assemble_matrix(k_g, lambda element: element.calculate_geometric_stiffness_matrix())

        # solve eigenvalue problem
        eigvals, eigvecs = eig(k_m, -k_g)

        # extract real parts of eigenvalues
        eigvals = np.array([x.real for x in eigvals])

        # sort eigenvalues and vectors
        idx = eigvals.argsort()
        eigvals = eigvals[idx]
        eigvecs = eigvecs[:, idx]

        # find index of closest eigenvalue to 1 (we could store all but that seems like an overkill)
        idx = (np.abs(eigvals - 1.0)).argmin()

        print('Closest eigenvalue: {}'.format(eigvals[idx]))
        print('Closest eigenvalue * lambda: {}'.format(eigvals[idx] * self.load_factor))  # this is printed in TRUSS
        if len(eigvecs[idx]) < 10:
            print('Closest eigenvector: {}'.format(eigvecs[idx]))

        self.first_eigenvalue = eigvals[idx]

        # store eigenvector as model
        model = self.get_duplicate()
        model._previous_model = self
        model.status = ModelStatus.eigenvector
        model.det_k = None
        model.first_eigenvalue = None
        model.first_eigenvector_model = None
        model.load_factor = None

        for index, dof in enumerate(assembler.dofs):
            model[dof].delta = eigvecs[index][idx]

        self.first_eigenvector_model = model

    def get_tangent_vector(self, assembler=None):
        """ Get the tangent vector

        Returns
        -------
        tangent : ndarray
            Tangent vector t = [v, 1]
            with v = d_u / d_lambda ... incremental velocity
        """
        if assembler is None:
            assembler = Assembler(self)

        dof_count = assembler.dof_count

        tangent = np.zeros(dof_count + 1)

        v = tangent[:-1]

        k = np.zeros((dof_count, dof_count))
        external_f = np.zeros(dof_count)

        # assemble stiffness
        assembler.assemble_matrix(k, lambda element: element.calculate_stiffness_matrix())

        # assemble force

        for i, dof in enumerate(assembler.dofs):
            external_f[i] += self[dof].external_force

        try:
            v[:dof_count] = la.solve(k, external_f)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        # lambda = 1
        tangent[-1] = 1

        return tangent

    # === prediction functions

    def predict_load_factor(self, value):
        """Predicts the solution by predictor_method lambda

        Parameters
        ----------
        value : float
            Value for the new load factor lambda.
        """
        self.status = ModelStatus.prediction
        self.load_factor = value

    def predict_load_increment(self, value):
        """Predicts the solution by incrementing lambda

        Parameters
        ----------
        value : float
            Value that is used to increment the load factor lambda.
        """
        self.status = ModelStatus.prediction
        self.load_factor += value

    def predict_dof_state(self, dof, value):
        """Predicts the solution by predictor_method the dof

        Parameters
        ----------
        dof : object
            Dof that is prescribed.
        value : float
            Value that is used to prescribe the dof.
        """
        self.status = ModelStatus.prediction
        self[dof].delta = value

    def predict_dof_increment(self, dof, value):
        """Predicts the solution by incrementing the dof

        Parameters
        ----------
        dof : object
            Dof that is incremented.
        value : float
            Value that is used to increment the dof.
        """
        self.status = ModelStatus.prediction
        self[dof].delta += value

    def predict_with_last_increment(self, value=None):
        """Predicts the solution by incrementing lambda and all dofs with the
           increment of the last solution step

        Parameters
        ----------
        value : float (optional)
            Length of the increment

        Raises
        ------
        RuntimeError
            If the model has not already one calculated step.
        """
        self.status = ModelStatus.prediction
        if self.get_previous_model().get_previous_model() is None:
            raise RuntimeError('predict_with_last_increment can only be used after the first step!')

        assembler = Assembler(self)

        last_increment = self.get_previous_model().get_increment_vector(assembler)

        length = la.norm(last_increment)

        if value is not None and length != 0.0:
            last_increment *= value/length
            length *= value

        if length == 0.0:
            print("WARNING: The length of the prescribed increment is 0.0!")

        # update dofs at model
        for index, dof in enumerate(assembler.dofs):
            self[dof].delta += last_increment[index]

        # update lam at model
        self.load_factor += last_increment[-1]

    def predict_tangential(self, strategy, **options):
        """ Make a tangential prediction

        Predicts the solution by incrementing lambda and all dofs with the
        increment of the last solution step.

        Parameters
        ----------
        strategy : str
            Strategy to scale the tangent vector.
            Avaiblable options:
            - 'lambda'
            - 'delta-lambda'
            - 'dof'
            - 'delta-dof'
            - 'arc-length'
        options
            value : float
                prescribed value according to the strategy
            dof : Object
                specifies the controlled dof for 'dof' and 'delta-dof' strategy
        """
        self.status = ModelStatus.prediction
        assembler = Assembler(self)

        # get tangent vector
        tangent = self.get_tangent_vector(assembler=assembler)

        # calculate scaling factor according to chosen strategy
        if strategy == 'lambda':
            prescribed_lam = options['value']

            delta_lambda = prescribed_lam - self.load_factor
            factor = delta_lambda

        elif strategy == 'delta-lambda':
            delta_lambda = options['value']

            factor = delta_lambda

        elif strategy == 'dof':
            dof = options['dof']
            value_prescribed = options['value']

            value = self[dof].delta

            delta_prescribed = value_prescribed - value

            dof_index = assembler.index_of_dof(dof)
            delta_current = tangent[dof_index]

            factor = delta_prescribed / delta_current

        elif strategy == 'delta-dof':
            dof = options['dof']
            delta_dof_prescribed = options['value']

            dof_index = assembler.index_of_dof(dof)
            delta_dof_current = tangent[dof_index]

            factor = delta_dof_prescribed / delta_dof_current

        elif strategy == 'arc-length':
            previous_model = self.get_previous_model()

            if previous_model.get_previous_model() is not None:
                previous_increment = previous_model.get_increment_vector(assembler)

            if 'value' in options.keys():
                prescribed_length = options['value']
            elif previous_model.get_previous_model() is not None:
                prescribed_length = la.norm(previous_increment)
            else:
                prescribed_length = 0.0

            if prescribed_length == 0.0:
                print("WARNING: The length of the prescribed increment is 0.0!")

            current_length = la.norm(tangent)

            factor = prescribed_length / current_length

            # tangent should point in a similar direction as the last increment
            if previous_model.get_previous_model() is not None:
                if previous_increment @ tangent < 0:
                    factor = -factor

        else:
            raise ValueError('Invalid strategy for prediction: {}'
                             .format(strategy))

        # scale tangent vector
        tangent *= factor

        # update dofs at model
        for index, dof in enumerate(assembler.dofs):
            self[dof].delta += tangent[index]

        # update lambda at model
        self.load_factor += tangent[-1]

    def combine_prediction_with_eigenvector(self, beta):
        """Combine the prediciton with the first eigenvector

        Parameters
        ----------
        beta : float
            factor between -1.0 and 1.0 used for a linear combination of the
            prediction with the eigenvector

        Raises
        ------
        RuntimeError
            If the model is not in prediction status
        ValueError
            If the beta is not between -1.0 and 1.0
        """
        if self.status != ModelStatus.prediction:
            raise RuntimeError('Model is not a predictor. Cannot combine with eigenvector!')

        if beta < -1.0 or beta > 1.0:
            raise ValueError('beta needs to be between -1.0 and 1.0')

        previous_model = self.get_previous_model()
        if previous_model.first_eigenvector_model is None:
            print('WARNING: solving eigenvalue problem in order to do branch switching')
            previous_model.solve_eigenvalues()

        eigenvector_model = previous_model.first_eigenvector_model

        assembler = Assembler(self)

        u_prediction = self.get_delta_dof_vector(model_b=previous_model, assembler=assembler)

        prediction_length = la.norm(u_prediction)

        eigenvector = eigenvector_model.get_delta_dof_vector(assembler=assembler)

        # scale eigenvector to the length of the prediction
        eigenvector *= (1.0/(la.norm(eigenvector)/prediction_length))

        prediction = u_prediction * (1.0 - abs(beta)) + eigenvector * beta

        delta_prediction = prediction - u_prediction

        # lambda = 0 for the eigenvector. Note: TRUSS.xls uses the same value as for the last increment
        delta_lam = - self.get_lam_increment()

        # update dofs at model
        for index, dof in enumerate(assembler.dofs):
            self[dof].delta += delta_prediction[index]

        # update lambda at model
        self.load_factor += delta_lam

    def scale_prediction(self, factor):
        """scale the prediction with a factor

        Parameters
        ----------
        factor : float
            factor used to scale the prediction


        Raises
        ------
        RuntimeError
            If the model is not in prediction status
        RuntimeError
            If the model has no previous model
        """
        if self.status != ModelStatus.prediction:
            raise RuntimeError('Model is not a predictor. Can only scale predictor!')

        if factor == 1.0:
            return

        previous_model = self.get_previous_model()

        if previous_model is None:
            raise RuntimeError('Previous Model is None!')

        assembler = Assembler(self)

        delta_dof_vector = self.get_delta_dof_vector(previous_model, assembler=assembler)

        delta_lambda = self.load_factor - previous_model.load_factor

        delta_dof_vector *= (factor - 1.0)
        delta_lambda *= (factor - 1.0)

        for i, dof in enumerate(assembler.dofs):
            self[dof].delta += delta_dof_vector[i]

        self.load_factor += delta_lambda

    def get_delta_dof_vector(self, model_b=None, assembler=None):
        """gets the delta dof between this and a given model_b as a numpy array

        Parameters
        ----------
        model_b : Model
            Model that is used as reference for the delta dof calculation. If
            not given, the initial model is used as reference.
        assembler: Assembler
            Assembler is used to order the dofs in the vector. If not given, a
            new assembler is created


        Returns
        ------
        delta : np.ndarray
            vector with the delta of all dofs

        Raises
        ------
        RuntimeError
            If the model is not in prediction status
        RuntimeError
            If the model has no previous model
        """
        if model_b is None:
            model_b = self.get_initial_model()

        if assembler is None:
            assembler = Assembler(self)

        delta = np.zeros(assembler.dof_count)

        for index, dof in enumerate(assembler.dofs):
            delta[index] = self[dof].delta - model_b[dof].delta

        return delta

    def load_displacement_curve(self, dof, skip_iterations=True):
        history = self.get_model_history(skip_iterations)

        data = np.zeros([2, len(history)])

        for i, self in enumerate(history):
            data[0, i] = self[dof].delta
            data[1, i] = self.load_factor

        return data

    def _repr_html_(self):
        from nfem.visualization.notebook_animation import show_animation
        return show_animation(self).data


class ModelStatus(Enum):
    """Enum for the model status """
    initial = 0
    duplicate = 1
    prediction = 2
    iteration = 3
    equilibrium = 4
    eigenvector = 5


class KeyCollection:
    def __init__(self):
        self._dictionary = OrderedDict()

    def __getitem__(self, key):
        if isinstance(key, int):
            return list(self._dictionary.values())[key]
        return self._dictionary[key]

    def _add(self, value):
        self._dictionary[value.id] = value

    def __contains__(self, key):
        return self._dictionary.__contains__(key)

    def __len__(self):
        return self._dictionary.values().__len__()

    def __iter__(self):
        return self._dictionary.values().__iter__()

    def __next__(self):
        return self._dictionary.values().__next__()

    def _ipython_key_completions_(self):
        return list(self._dictionary.keys())



"""This module contains the Newton-Raphson algorithm.

Author: Armin Geiser
"""

import numpy as np
import numpy.linalg as la

def newton_raphson_solve(calculate_system, x_initial, max_iterations=100, tolerance=1e-7):
    """Solves the nonlinear system defined by the `calculate_system` callback.

    The array with the initial solution is updated during the solve and contains
    the solution at convergence.

    Parameters
    ----------
    calculate_system : function        
        This function is called several times to evaluate the function (rhs) and the 
        functions derivatives (lhs) with a given state (x)

        It should look like this:

        def calculate_system(x):
            ...
            return lhs, rhs
    x_initial : ndarray
        Initial guess of the solution
    max_iterations : int
        Maximum number of iterations
    tolerance : float
        Convergence tolerance value for the residual norm

    Raises
    ----------
    RuntimeError
        If the algorithm does not converge within `max_iterations`        
    """
    x = x_initial
    residual_norm = None

    for i in range(1, max_iterations + 1):

        # calculate left and right hand side
        lhs, rhs = calculate_system(x)

        # calculate residual
        residual_norm = la.norm(rhs)

        # check convergence
        if residual_norm < tolerance:
            print('  Newthon-Raphson converged in step {}.'.format(i))
            print('  Residual norm: {}.'.format(residual_norm))
            return x, i

        # compute delta_x

        try:
            delta_x = la.solve(lhs, rhs)
        except np.linalg.LinAlgError:
            raise RuntimeError('Stiffness matrix is singular')

        # update x
        x -= delta_x

    raise RuntimeError('Newthon-Raphson did not converge after {} steps. Residual norm: {}'
                       .format(max_iterations, residual_norm))



from matplotlib.patches import FancyArrowPatch, Circle
from mpl_toolkits.mplot3d import proj3d

import numpy as np


class Arrow3D(FancyArrowPatch):
    """from https://stackoverflow.com/questions/22867620/putting-arrowheads-on-vectors-in-matplotlibs-3d-plot"""
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


class Circle3D(Circle):
    def __init__(self, x, y, z, r, *args, **kwargs):
        Circle.__init__(self, (0, 0), r, *args, **kwargs)
        self._verts3d = x, y, z

    def draw(self, renderer):
        x3d, y3d, z3d = self._verts3d
        x, y, z = proj3d.proj_transform(x3d, y3d, z3d, renderer.M)
        self.center = x, y
        Circle.draw(self, renderer)


def get_force_arrow(x, y, z, fx, fy, fz, length, *args, **kwargs):
    delta = np.array([fx, fy, fz])
    dnorm = np.linalg.norm(delta)
    if dnorm < 1e-8:
        return None
    delta = delta * length / dnorm
    return Arrow3D([x, x-delta[0]], [y, y-delta[1]], [z, z-delta[2]], *args, mutation_scale=15, lw=1.5, arrowstyle="<|-", **kwargs)


def get_dof_arrow(x, y, z, dx, dy, dz, length, *args, **kwargs):
    delta = np.array([dx, dy, dz])
    delta = delta/np.linalg.norm(delta)*length
    return Arrow3D([x, x+delta[0]], [y, y+delta[1]], [z, z+delta[2]], *args, mutation_scale=15, lw=1.5, arrowstyle="-|>", **kwargs)


def get_sphere(x, y, z, r, *args, **kwargs):
    return Circle3D(x, y, z, r, *args, **kwargs)


def get_tet4_polygons(x, y, z, h, direction):
    d = h/2
    node_1 = (x, y, z)
    if direction == 'u':
        node_2 = (x-h, y-d, z-d)
        node_3 = (x-h, y+d, z-d)
        node_4 = (x-h, y+d, z+d)
        node_5 = (x-h, y-d, z+d)
    elif direction == 'v':
        node_2 = (x-d, y-h, z-d)
        node_3 = (x-d, y-h, z+d)
        node_4 = (x+d, y-h, z+d)
        node_5 = (x+d, y-h, z-d)
    elif direction == 'w':
        node_2 = (x-d, y-d, z-h)
        node_3 = (x-d, y+d, z-h)
        node_4 = (x+d, y+d, z-h)
        node_5 = (x+d, y-d, z-h)

    poly_1 = [node_1, node_2, node_3]
    poly_2 = [node_1, node_3, node_4]
    poly_3 = [node_1, node_3, node_5]
    poly_4 = [node_1, node_5, node_2]
    poly_5 = [node_2, node_3, node_4, node_5]

    return poly_1, poly_2, poly_3, poly_4, poly_5



# PythonUI by Thomas Oberbichler
# https://github.com/oberbichler/PythonUI

import matplotlib
matplotlib.use("Qt5Agg") 
from matplotlib.backends.backend_qt5agg import (FigureCanvasQTAgg,
                                                NavigationToolbar2QT)
from matplotlib.figure import Figure
from PyQt5 import Qt, QtCore, QtGui, QtWidgets
import inspect
import numpy as np
import sys


class _GenericValidator(QtGui.QValidator):
    def __init__(self, parent, validate):
        super(_GenericValidator, self).__init__(parent)
        self._validate = validate

    def validate(self, string, pos):
        validated = self._validate(string)

        if validated is None:
            return QtGui.QValidator.Invalid, string, pos
        elif validated == string:
            return QtGui.QValidator.Acceptable, string, pos
        else:
            return QtGui.QValidator.Intermediate, string, pos

    def fixup(self, string):
        return self._validate(string)


class Option(QtCore.QObject):
    _changed = QtCore.pyqtSignal

    def __init__(self, value, action=None):
        super(Option, self).__init__()
        self.value = value

        if action:
            self.connect(action)

    def connect(self, action):
        self._changed.connect(action)

    def change(self, value):
        self.value = value

    def __call__(self):
        return self.value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        self.emit()

    def emit(self):
        self._changed.emit(self._value)


class Fore:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    RESET = '\033[39m'


class Back:
    BLACK = '\033[40m'
    RED = '\033[41m'
    GREEN = '\033[42m'
    YELLOW = '\033[43m'
    BLUE = '\033[44m'
    MAGENTA = '\033[45m'
    CYAN = '\033[46m'
    WHITE = '\033[47m'
    RESET = '\033[49m'


class Style:
    DIM = '\033[2m'
    NORMAL = '\033[22m'
    BRIGHT = '\033[1m'
    RESET_ALL = '\033[0m'


ConsoleStyles = Fore, Back, Style


class DarkStyle:
    BACKGROUND = QtGui.QColor('#1b1b1b')
    BLACK = QtGui.QColor('#303030')
    RED = QtGui.QColor('#e1321a')
    GREEN = QtGui.QColor('#6ab017')
    YELLOW = QtGui.QColor('#ffc005')
    BLUE = QtGui.QColor('#004f9e')
    MAGENTA = QtGui.QColor('#ec0048')
    CYAN = QtGui.QColor('#2aa7e7')
    WHITE = QtGui.QColor('#f2f2f2')


class Console(QtWidgets.QTextEdit):

    def __init__(self):
        super(Console, self).__init__()
        
        font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        
        self.style = DarkStyle

        font = QtGui.QFont('Consolas')
        font.setStyleHint(QtGui.QFont.TypeWriter)
        self.setFont(font)
        self.setReadOnly(True)
        self.setFrameStyle(QtWidgets.QFrame.NoFrame)
        self._default_format = QtGui.QTextCharFormat()
        self._text_format = QtGui.QTextCharFormat()
        self.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)

        p = self.palette()
        p.setColor(QtGui.QPalette.Base, self.style.BACKGROUND)
        p.setColor(QtGui.QPalette.Text, self.style.WHITE)
        self.setPalette(p)

    def _foreground(self, color):
        if color is None:
            color = self._default_format.foreground()
        self._text_format.setForeground(color)

    def _background(self, color):
        if color is None:
            color = self._default_format.background()
        self._text_format.setBackground(color)

    def _apply_code(self, code):
        if code == 0:
            self._text_format.setFontWeight(self._text_format.fontWeight())
            self._foreground(None)
            self._background(None)
        elif code == 1:
            self._text_format.setFontWeight(QtGui.QFont.Bold)
        elif code == 30:
            self._foreground(self.style.BLACK)
        elif code == 31:
            self._foreground(self.style.RED)
        elif code == 32:
            self._foreground(self.style.GREEN)
        elif code == 33:
            self._foreground(self.style.YELLOW)
        elif code == 34:
            self._foreground(self.style.BLUE)
        elif code == 35:
            self._foreground(self.style.MAGENTA)
        elif code == 36:
            self._foreground(self.style.CYAN)
        elif code == 37:
            self._foreground(self.style.WHITE)
        elif code == 39:
            self._foreground(None)
        elif code == 40:
            self._background(self.style.BLACK)
        elif code == 41:
            self._background(self.style.RED)
        elif code == 42:
            self._background(self.style.GREEN)
        elif code == 43:
            self._background(self.style.YELLOW)
        elif code == 44:
            self._background(self.style.BLUE)
        elif code == 45:
            self._background(self.style.MAGENTA)
        elif code == 46:
            self._background(self.style.CYAN)
        elif code == 47:
            self._background(self.style.WHITE)
        elif code == 49:
            self._background(None)

    def write(self, text):
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)

        npos = 0

        while True:
            start = text.find('\033[', npos)

            if start < 0:
                cursor.insertText(text[npos:], self._text_format)
                break

            if start != npos:
                cursor.insertText(text[npos:start], self._text_format)

            end = text.find('m', start + 2)

            for code in map(int, text[start+2:end].split(';')):
                self._apply_code(code)

            npos = end + 1

        self.setTextCursor(cursor)
        self.ensureCursorVisible()

    def flush(self):
        pass


class WidgetBuilder:
    def __init__(self, ground, context):
        self._ground = ground
        self.context = context

    def _add_widget(self, widget):
        self._ground.addWidget(widget)

    def add(self, widget_type):
        widget = widget_type()
        builder = WidgetBuilder(self._ground, self.context)
        widget.build(builder)
        self._add_widget(widget)

    def add_space(self):
        spacer_item = QtWidgets.QSpacerItem(16, 16)
        self._ground.addItem(spacer_item)

    def add_stretch(self):
        self._ground.addStretch(1)

    def add_label(self, label):
        label_widget = QtWidgets.QLabel(label)
        self._add_widget(label_widget)

    def add_button(self, label, action):
        button_widget = QtWidgets.QPushButton(label)
        self._add_widget(button_widget)

        if len(inspect.signature(action).parameters) == 0:
            button_widget.clicked.connect(action)
        else:
            button_widget.clicked.connect(lambda: action(self.context))

    def add_textbox(self, label, option, prefix=None, postfix=None,
                    validate=None, readonly=False):
        if label:
            label_widget = QtWidgets.QLabel(label)
            self._add_widget(label_widget)

        row_widget = QtWidgets.QWidget()
        row_layout = QtWidgets.QHBoxLayout()
        row_layout.setContentsMargins(0, 0, 0, 0)
        row_widget.setLayout(row_layout)
        self._add_widget(row_widget)

        if prefix:
            prefix_widget = QtWidgets.QLabel(prefix)
            row_layout.addWidget(prefix_widget)

        textbox_widget = QtWidgets.QLineEdit()
        if validate:
            validator = _GenericValidator(textbox_widget, validate)
            textbox_widget.setValidator(validator)
        textbox_widget.setText(str(option.value))
        textbox_widget.setReadOnly(readonly)

        row_layout.addWidget(textbox_widget, 1)

        if option:
            option.connect(lambda value: textbox_widget.setText(str(value)))
            textbox_widget.editingFinished.connect(
                lambda: option.change(textbox_widget.text()))

        if postfix:
            postfix_widget = QtWidgets.QLabel(postfix)
            row_layout.addWidget(postfix_widget)

    def add_spinbox(self, label, option, prefix=None, postfix=None,
                    dtype=float, minimum=None, maximum=None, step=None,
                    decimals=None):
        if label:
            label_widget = QtWidgets.QLabel(label)
            self._add_widget(label_widget)

        row_widget = QtWidgets.QWidget()
        row_layout = QtWidgets.QHBoxLayout()
        row_layout.setContentsMargins(0, 0, 0, 0)
        row_widget.setLayout(row_layout)
        self._add_widget(row_widget)

        if prefix:
            prefix_widget = QtWidgets.QLabel(prefix)
            row_layout.addWidget(prefix_widget)

        if dtype is int:
            spinbox_widget = QtWidgets.QSpinBox()
            spinbox_widget.setMinimum(minimum or -2147483648)
            spinbox_widget.setMaximum(maximum or 2147483647)
            spinbox_widget.setSingleStep(step or 1)
            spinbox_widget.setValue(option.value)
        elif dtype is float:
            spinbox_widget = QtWidgets.QDoubleSpinBox()
            spinbox_widget.setMinimum(minimum or -Qt.qInf())
            spinbox_widget.setMaximum(maximum or Qt.qInf())
            spinbox_widget.setSingleStep(step or 0.1)
            spinbox_widget.setDecimals(decimals or 5)
            spinbox_widget.setValue(option.value)
        else:
            raise ValueError(f'Invalid dtype "{dtype.__name__}"')

        spinbox_widget.setKeyboardTracking(False)
        row_layout.addWidget(spinbox_widget, 1)

        if option:
            option.connect(spinbox_widget.setValue)
            spinbox_widget.valueChanged.connect(option.change)

        if postfix:
            postfix_widget = QtWidgets.QLabel(postfix)
            row_layout.addWidget(postfix_widget)

    def add_checkbox(self, label, option):
        checkbox_widget = QtWidgets.QCheckBox(label)
        checkbox_widget.setChecked(option.value)
        self._add_widget(checkbox_widget)

        option.connect(checkbox_widget.setChecked)
        checkbox_widget.clicked.connect(option.change)

    def add_combobox(self, items, option, label=None):
        if label:
            label_widget = QtWidgets.QLabel(label)
            self._add_widget(label_widget)

        combobox_widget = QtWidgets.QComboBox()
        self._add_widget(combobox_widget)

        for item in items:
            combobox_widget.addItem(str(item))

        if option:
            option.connect(combobox_widget.setCurrentIndex)
            combobox_widget.currentIndexChanged.connect(option.change)
            combobox_widget.setCurrentIndex(option.value)

    def add_radiobuttons(self, items, option):
        button_group = QtWidgets.QButtonGroup(self._ground.parent())

        for i, item in enumerate(items):
            radio_button = QtWidgets.QRadioButton(item)
            button_group.addButton(radio_button, i)
            self._add_widget(radio_button)

        def toggled(button, checked):
            if not checked:
                return
            button_id = button_group.id(button)
            option.change(button_id)

        def toggle(button_id):
            button = button_group.button(button_id)
            button.toggle()

        if option:
            toggle(option.value)
            button_group.buttonToggled.connect(toggled)
            option.connect(toggle)

    def add_group(self, label, content):
        group_widget = QtWidgets.QGroupBox(label)
        self._add_widget(group_widget)

        group_layout = QtWidgets.QVBoxLayout()
        group_widget.setLayout(group_layout)

        content_widget = content()
        content_widget._build(self.context)

        group_layout.addWidget(content_widget)

    def add_tabs(self, items, option=None):
        tabs_widget = TabsWidget(self.context)
        self._add_widget(tabs_widget)

        for label, widget_type in items:
            tabs_widget.add_tab(label, widget_type)

        if option:
            option.connect(tabs_widget.setCurrentIndex)
            tabs_widget.currentChanged.connect(option.change)

    def add_stack(self, items, option=None):
        stack_widget = StackWidget(self.context, items, option)
        self._add_widget(stack_widget)

    def add_pages(self, items, option=None):
        pages_widget = PagesWidget(self.context, items, option)
        self._add_widget(pages_widget)

    def add_array(self, option, label=None, readonly=False):
        if label:
            label_widget = QtWidgets.QLabel(label)
            self._add_widget(label_widget)

        table_widget = QtWidgets.QTableWidget()
        if readonly:
            table_widget.setEditTriggers(
                QtWidgets.QAbstractItemView.NoEditTriggers)

        def update_table(array):
            table_widget.blockSignals(True)

            shape = np.shape(array)

            if len(shape) == 1:
                rows, = shape
                table_widget.setRowCount(rows)
                table_widget.setColumnCount(1)

                table_widget.setVerticalHeaderLabels(map(str, range(rows)))
                table_widget.setHorizontalHeaderLabels(['0'])

                for (i,), value in np.ndenumerate(array):
                    value_str = str(value)
                    table_widget.setItem(i, 0,
                                         QtWidgets.QTableWidgetItem(value_str))
            elif len(shape) == 2:
                rows, cols = shape
                table_widget.setRowCount(rows)
                table_widget.setColumnCount(cols)

                table_widget.setVerticalHeaderLabels(map(str, range(rows)))
                table_widget.setHorizontalHeaderLabels(map(str, range(cols)))

                for (i, j), value in np.ndenumerate(array):
                    value_str = str(value)
                    table_widget.setItem(i, j,
                                         QtWidgets.QTableWidgetItem(value_str))
            else:
                raise Exception('Arrays with dimension > 2 not supported')

            table_widget.blockSignals(False)

        def update_cell(row, col):
            shape = np.shape(option.value)

            value = float(table_widget.item(row, col).text())

            if len(shape) == 1:
                option.value[row] = value
            elif len(shape) == 2:
                option.value[row, col] = value
            else:
                raise Exception('Arrays with dimension > 2 not supported')

            option.emit()

        update_table(option.value)

        option.connect(update_table)
        table_widget.cellChanged.connect(update_cell)

        self._add_widget(table_widget)

    """
    contributed by Mahmoud Zidan
    """
    def add_slider(self, label, option, minimum=None, maximum=None, ticks=None):
        if label:
            label_widget = QtWidgets.QLabel(label)
            self._add_widget(label_widget)

        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider.setMinimum(minimum or 1)
        slider.setMaximum(maximum or 10)
        if ticks:
            slider.setTickInterval(ticks)
            slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        slider.setValue(option.value)
        slider.valueChanged.connect(option.change)
        self._add_widget(slider)

    def add_wheel(self, option, unit=1):
        wheel = Wheel(option, unit)
        self._add_widget(wheel)

class Widget(QtWidgets.QWidget):
    def __init__(self):
        super(Widget, self).__init__()

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._ground = layout

    def _build(self, context):
        builder = WidgetBuilder(self._ground, context)
        self.build(builder)

    def build(self, builder):
        pass


class StackWidget(QtWidgets.QWidget):
    def __init__(self, context, items, option):
        super(StackWidget, self).__init__()

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        stack = QtWidgets.QStackedWidget()
        layout.addWidget(stack)
        self._stack = stack

        for widget_type in items:
            widget = widget_type()

            builder = WidgetBuilder(widget._ground, context)
            widget.build(builder)

            if stack.count() != 0:
                widget.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                     QtWidgets.QSizePolicy.Ignored)

            stack.addWidget(widget)

        if option:
            option.connect(self._select_index)

        self._select_index(0)

    def _select_index(self, index):
        if self._stack.currentWidget():
            self._stack.currentWidget().setSizePolicy(
                QtWidgets.QSizePolicy.Ignored,
                QtWidgets.QSizePolicy.Ignored)

        self._stack.setCurrentIndex(index)

        if self._stack.currentWidget():
            self._stack.currentWidget().setSizePolicy(
                QtWidgets.QSizePolicy.Preferred,
                QtWidgets.QSizePolicy.Preferred)


class PagesWidget(QtWidgets.QWidget):
    def __init__(self, context, pages, option):
        super(PagesWidget, self).__init__()

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        combobox = QtWidgets.QComboBox()
        layout.addWidget(combobox)
        self._combobox = combobox

        stack = QtWidgets.QStackedWidget()
        layout.addWidget(stack)
        self._stack = stack

        for label, widget_type in pages:
            content = widget_type()

            builder = WidgetBuilder(content._ground, context)
            content.build(builder)

            self._add_page(label, content)

        combobox.currentIndexChanged.connect(self._select_index)

        if option:
            combobox.currentIndexChanged.connect(option.change)
            option.connect(self._select_index)

        self._select_index(0)

    def _select_index(self, index):
        self._combobox.setCurrentIndex(index)

        if self._stack.currentWidget():
            self._stack.currentWidget().setSizePolicy(
                QtWidgets.QSizePolicy.Ignored,
                QtWidgets.QSizePolicy.Ignored)

        self._stack.setCurrentIndex(index)

        if self._stack.currentWidget():
            self._stack.currentWidget().setSizePolicy(
                QtWidgets.QSizePolicy.Preferred,
                QtWidgets.QSizePolicy.Preferred)

    def _add_page(self, key, widget):
        if self._stack.count() != 0:
            widget.setSizePolicy(
                QtWidgets.QSizePolicy.Ignored,
                QtWidgets.QSizePolicy.Ignored)

        self._combobox.addItem(str(key), key)

        self._stack.addWidget(widget)

        return widget


class TabsWidget(QtWidgets.QTabWidget):
    def __init__(self, context):
        super(TabsWidget, self).__init__()
        self.context = context

    def add_tab(self, label, widget_type=None):
        widget = widget_type()

        builder = WidgetBuilder(widget._ground, self.context)
        widget.build(builder)

        widget.setContentsMargins(8, 8, 8, 8)

        self.addTab(widget, label)

        return widget


class Sidebar(QtWidgets.QScrollArea):
    def __init__(self):
        super(Sidebar, self).__init__()

        widget = Widget()

        self._ground = widget._ground

        self.setWidget(widget)
        self._ground.setContentsMargins(8, 8, 8, 8)
        self.horizontalScrollBar().setEnabled(False)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setWidgetResizable(True)
        self.setMinimumWidth(350)


class PlotCanvas(QtWidgets.QWidget):
    def __init__(self, parent):
        super(PlotCanvas, self).__init__()

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)

        figure = Figure()
        canvas = FigureCanvasQTAgg(figure)
        canvas.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(canvas, 1, 1, 1, 1)
        self._canvas = canvas

        toolbar = NavigationToolbar2QT(canvas, self)
        layout.addWidget(toolbar, 2, 1, 1, 1)

        plot = figure.add_subplot(111)
        plot.set_aspect('equal')
        self._plot = plot

    def redraw(self):
        plot = self._plot

        figure = plot.get_figure()

        for ax in figure.axes[1:]:
            figure.delaxes(ax)

        figure.subplots_adjust()

        plot.clear()

        self.plot(plot)

    def plot(self, ax):
        pass


class Wheel(QtWidgets.QDial):
    def __init__(self, option, unit=1):
        super().__init__()
        self.setMaximum(359)
        self.setValue(180)
        self._value = 180
        self.unit = unit
        self.valueChanged.connect(self._valueChanged)            
        self.setWrapping(True)
        self.option = option

    def _valueChanged(self, value):
        old_angle = self._value / 180 * np.pi
        new_angle = value / 180 * np.pi

        delta = np.arccos(np.cos(old_angle - new_angle))
        s = np.sign(-np.sin(old_angle - new_angle))

        self._value = int(self._value + s * delta * 180 / np.pi)

        self.option.value = self._value / 180 * self.unit


class ApplicationWindow(QtWidgets.QWidget):
    def __init__(self, title='', size=(1200, 800), content=None):
        super(ApplicationWindow, self).__init__()

        self.resize(*size)
        self.setWindowTitle(title)

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)

        self.content = content(parent=self) if content else None
        self.console = Console()
        self.sidebar = Sidebar()

        if content:
            vsplitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
            vsplitter.addWidget(self.content)
            vsplitter.addWidget(self.console)
            vsplitter.setStretchFactor(0, 1)
            vsplitter.setStretchFactor(1, 0)

        hsplitter = QtWidgets.QSplitter()
        hsplitter.addWidget(self.sidebar)
        if content:
            hsplitter.addWidget(vsplitter)
        else:
            hsplitter.addWidget(self.console)
        hsplitter.setStretchFactor(0, 0)
        hsplitter.setStretchFactor(1, 1)

        self.layout().addWidget(hsplitter)

    def closeEvent(self, event):
        QtWidgets.QApplication.quit()

    @classmethod
    def run(cls, *args, **kwargs):
        app = QtWidgets.QApplication([])

        app.setStyle(QtWidgets.QStyleFactory.create('Fusion'))

        widget = cls(*args, **kwargs)
        widget._build(widget)

        widget.show()

        widget._started()

        app.exec_()

        return widget

    def show_dialog(self, widget_type, title='', size=None, modal=True,
                    action=None):
        dialog = QtWidgets.QDialog(self, QtCore.Qt.Tool)
        dialog.setWindowTitle(title)
        dialog.setModal(modal)

        widget = widget_type()
        widget._build(self)

        if size:
            dialog.resize(*size)
        else:
            dialog.adjustSize()

        layout = QtWidgets.QGridLayout()
        dialog.setLayout(layout)
        layout.addWidget(widget)

        def close_dialog(message=None):
            dialog.close()
            if action:
                action(message)

        self.close_dialog = close_dialog

        dialog.show()

    def show_open_file_dialog(self, title=None, extension_filters=None):
        if isinstance(extension_filters, list):
            extension_filter = ';;'.join(extension_filters)
        else:
            extension_filter = extension_filters

        result = QtWidgets.QFileDialog.getOpenFileName(self, title,
                                                       filter=extension_filter)

        return result

    def show_save_file_dialog(self, title=None, extension_filters=None):
        if isinstance(extension_filters, list):
            extension_filter = ';;'.join(extension_filters)
        else:
            extension_filter = extension_filters

        result = QtWidgets.QFileDialog.getSaveFileName(self, title,
                                                       filter=extension_filter)

        return result

    def show_error_dialog(self, message):
        QtWidgets.QMessageBox.critical(self, 'Error', message)

    def showEvent(self, event):
        self._old_stdout = sys.stdout
        sys.stdout = self.console

    def hideEvent(self, event):
        sys.stdout = self._old_stdout

    def _build(self, context):
        sidebar_builder = WidgetBuilder(self.sidebar._ground, context)
        self._build_sidebar(sidebar_builder)

    def _build_sidebar(self, builder):
        pass

    def _started(self):
        pass



"""This module only contains the spring element.

Authors: Thomas Oberbichler
"""

import numpy as np
import numpy.linalg as la


class Spring:
    def __init__(self, id, node, kx=0, ky=0, kz=0):
        self.id = id
        self.node = node
        self.kx = kx
        self.ky = ky
        self.kz = kz

    @property
    def dofs(self):
        node = self.node
        return [node._dof_x, node._dof_y, node._dof_z]

    def calculate_elastic_stiffness_matrix(self):
        return np.array([[self.kx, 0, 0], [0, self.ky, 0], [0, 0, self.kz]])

    def calculate_material_stiffness_matrix(self):
        return None

    def calculate_initial_displacement_stiffness_matrix(self):
        return None

    def calculate_geometric_stiffness_matrix(self, linear=False):
        return None

    def calculate_stiffness_matrix(self):
        return self.calculate_elastic_stiffness_matrix()

    def calculate_internal_forces(self):
        return np.array([
            self.kx * self.node.u,
            self.ky * self.node.v,
            self.kz * self.node.w,
        ])



"""This module contains helpers for visualize data.

Authors: Klaus Sautter, Thomas Oberbichler, Armin Geiser
"""

import numpy as np
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection
import matplotlib.pyplot as plt
import matplotlib.animation as anim

from nfem.visualization.plot_symbols import get_force_arrow, get_tet4_polygons, get_dof_arrow

from nfem.truss import Truss
from nfem.model import ModelStatus


class Plot2D:
    def __init__(self, x_label='Displacement', y_label=r'Load factor ($\lambda$)',
                 title='Load-displacement diagram'):
        self.fig, self.ax = plt.subplots()

        self.ax.set(xlabel=x_label, ylabel=y_label, title=title)
        self.ax.set_facecolor('white')
        self.ax.grid()

        self.legend = []

    def invert_xaxis(self):
        """Invert the x axis of the plot"""
        self.ax.invert_xaxis()

    def add_load_displacement_curve(self, model, dof, label=None, show_iterations=False):
        plot_load_displacement_curve(self.ax, model, dof, label)
        if show_iterations:
            plot_load_displacement_iterations(self.ax, model, dof, label)

    def add_det_k_curve(self, model, dof, label=None):
        plot_det_k_curve(self.ax, model, dof, label=label)

    def add_history_curve(self, model, x_y_data, fmt='-o', **kwargs):
        """Add a history curve of the model to the plot.

        Parameters
        ----------
        model : Model
            Model object of which the history will be printed
        x_y_data : function(Model) that returns the value for the x and y axis
            of a models state. It is called for all models in the history
        fmt: matplotlib format string e.g. '-o' for line with points
            For details visit the link below
        **kwargs: additional format arguments e.g. label="My label" to give the
            curve a name.
            for details visit the link below
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html
        """
        plot_history_curve(self.ax, model, x_y_data, fmt, **kwargs)

    def add_custom_curve(self, *args, **kwargs):
        """Add a custom curve to the plot.
        Uses the syntax of the matplotlip.pyplot.plot function

        For a description of the possible parameters visit
        https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html
        """
        plot_custom_curve(self.ax, *args, **kwargs)

    def show(self, block=True):
        """Shows the plot with all the curves that have been added.
        """
        self.ax.legend(loc='best')
        plt.show(block=block)


class Animation3D:
    def show(self, model, speed=200):
        print("WARNING: Please use the function 'show_animation' instead of this class!")
        self.animation = show_animation(model, speed)


class DeformationPlot3D:
    def show(self, model, step=None):
        print("WARNING: Please use the function 'show_deformation_plot' instead of this class!")
        show_deformation_plot(model, step)


def get_bounding_box(models):
    nodes = [node for model in models for node in model.nodes]

    min_x = min(node.x for node in nodes)
    max_x = max(node.x for node in nodes)

    min_y = min(node.y for node in nodes)
    max_y = max(node.y for node in nodes)

    min_z = min(node.z for node in nodes)
    max_z = max(node.z for node in nodes)

    return min_x, max_x, min_y, max_y, min_z, max_z


def plot_scaled_model(ax, model, color, **options):
    scaling_factor = options.get('plot/scaling_factor', None)
    if not scaling_factor:
        # autoscaling max u to 10% of bounding_box
        bounding_box = get_bounding_box([model.get_initial_model()])
        min_x, max_x, min_y, max_y, min_z, max_z = bounding_box
        max_delta = max(max_x-min_x, max_y-min_y, max_z-min_z)
        max_u = max(abs(node.u) for node in model.nodes)
        max_v = max(abs(node.v) for node in model.nodes)
        max_w = max(abs(node.w) for node in model.nodes)
        max_def = max(max_u, max_v, max_w)
        scaling_factor = max_delta / max_def * 0.1

    lines = list()

    for element in model.elements:
        if type(element) == Truss:
            node_a = element.node_a
            node_b = element.node_b

            b = [node_b.reference_x+scaling_factor*node_b.u,
                 node_b.reference_y+scaling_factor*node_b.v,
                 node_b.reference_z+scaling_factor*node_b.w]
            a = [node_a.reference_x+scaling_factor*node_a.u,
                 node_a.reference_y+scaling_factor*node_a.v,
                 node_a.reference_z+scaling_factor*node_a.w]

            lines.append([a, b])

    lc = Line3DCollection(lines, colors=color, linewidths=2)

    ax.add_collection(lc)

    plot_symbols(ax, model, color, initial=False, **options)


def plot_model(ax, model, color, initial, **options):
    lines = list()

    for element in model.elements:
        if type(element) == Truss:
            node_a = element.node_a
            node_b = element.node_b

            a = [node_a.reference_x, node_a.reference_y, node_a.reference_z] if initial else [node_a.x, node_a.y, node_a.z]
            b = [node_b.reference_x, node_b.reference_y, node_b.reference_z] if initial else [node_b.x, node_b.y, node_b.z]

            lines.append([a, b])

    lc = Line3DCollection(lines, colors=color, linewidths=2)

    ax.add_collection(lc)

    plot_symbols(ax, model, color, initial, **options)


def plot_symbols(ax, model, color, initial, **options):
    if options.get('plot/dirichlet', True):
        plot_boundary_conditions(ax, model, initial, **options)
    if options.get('plot/neumann', True):
        plot_forces(ax, model, initial, **options)
    if options.get('plot/highlight_dof', False):
        plot_dof_higlight(ax, model, initial, **options)


def get_max_axes_delta(ax):
    x_lim = ax.get_xlim()
    y_lim = ax.get_ylim()
    z_lim = ax.get_zlim()
    return max([x_lim[1]-x_lim[0], y_lim[1]-y_lim[0], z_lim[1]-z_lim[0]])


def plot_dof_higlight(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/25 * options.get('plot/symbol_size', 5)

    dof = model.dofs[options.get('plot/dof_idx', None)]
    if dof is None:
        return

    node_id, dof_type = dof

    node = model.nodes[node_id]

    dx, dy, dz = 0, 0, 0
    if dof_type == 'u':
        dx = 1
    if dof_type == 'v':
        dy = 1
    if dof_type == 'w':
        dz = 1

    color = 'lightgray' if initial else 'tab:blue'
    if initial:
        x = node.reference_x
        y = node.reference_y
        z = node.reference_z
    else:
        x = node.x
        y = node.y
        z = node.z
    a = get_dof_arrow(x, y, z, dx, dy, dz, size*0.75, color=color)
    ax.add_artist(a)
    # TODO fix size of spere...
    # a = get_sphere(x, y, z, size/300, color=color)
    # ax.add_artist(a)


def plot_forces(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/25 * options.get('plot/symbol_size', 5)

    for node in model.nodes:
        color = 'lightgray' if initial else 'lightcoral'
        if initial:
            x, y, z = node.reference_location
        else:
            x, y, z = node.location
        a = get_force_arrow(x, y, z, node.fx, node.fy, node.fz, size, color=color)

        if a is not None:
            ax.add_artist(a)


def plot_boundary_conditions(ax, model, initial, **options):
    size = get_max_axes_delta(ax)/100.0 * options.get('plot/symbol_size', 5)

    polygons = list()

    for node in model.nodes:
        for dof_type in ['u', 'v', 'w']:
            if node.dof(dof_type).is_active:
                continue
            if initial:
                polygons.extend(get_tet4_polygons(node.reference_x, node.reference_y, node.reference_z, size, dof_type))
            else:
                polygons.extend(get_tet4_polygons(node.x, node.y, node.z, size, dof_type))

    color = 'lightgray' if initial else 'lightcoral'
    pc = Poly3DCollection(polygons, edgecolor=color, linewidth=0.5, alpha=0.25)
    pc.set_facecolor(color)  # needs to be defined outside otherwhise alpha is not working
    ax.add_collection3d(pc)


def animate_model(fig, ax, models, speed=200, **options):
    bounding_box = get_bounding_box(models)

    def update(step):
        step_model = models[step]

        ax.clear()

        ax.grid()

        plot_bounding_cube(ax, bounding_box)

        ax.set_xlabel('< x >')
        ax.set_ylabel('< y >')
        ax.set_zlabel('< z >')

        ax.set_title('Deformed structure at time step {}\n{}'.format(step, step_model.name))

        plot_model(ax, step_model, 'gray', True, **options)
        plot_model(ax, step_model, 'red', False, **options)

    a = anim.FuncAnimation(fig, update, frames=len(models), repeat=True, interval=speed)

    return a


def plot_load_displacement_iterations(ax, model, dof, label=None):
    history = model.get_model_history(skip_iterations=False)

    x_data = np.zeros(len(history))
    y_data = np.zeros(len(history))

    node_id, dof_type = dof

    for i, model in enumerate(history):
        x_data[i] = model[dof].delta
        y_data[i] = model.load_factor

    if label is None:
        label = r'$\lambda$ : {} at node {} (iter)'.format(dof_type, node_id)
    else:
        label += ' (iter)'
    ax.plot(x_data, y_data, '--o', linewidth=0.75, markersize=2.0, label=label)


def plot_load_displacement_curve(ax, model, dof, label=None):
    history = model.get_model_history()

    x_data = np.zeros(len(history))
    y_data = np.zeros(len(history))

    node_id, dof_type = dof

    for i, model in enumerate(history):
        x_data[i] = model[dof].delta
        y_data[i] = model.load_factor

    if label is None:
        label = r'$\lambda$ : {} at node {}'.format(dof_type, node_id)
    ax.plot(x_data, y_data, '-o', label=label)


def plot_det_k_curve(ax, model, dof, label=None):
    history = model.get_model_history()

    x_data = np.zeros(len(history))
    y_data = np.zeros(len(history))

    node_id, dof_type = dof

    for i, model in enumerate(history):
        x_data[i] = model[dof].delta
        y_data[i] = model.det_k

    if label is None:
        label = 'det(K) : {} at node {}'.format(dof_type, node_id)
    ax.plot(x_data, y_data, '-o', label=label)


def plot_history_curve(ax, model, xy_function, fmt, skip_iterations=True, **kwargs):
    history = model.get_model_history(skip_iterations)

    x_data = np.zeros(len(history))
    y_data = np.zeros(len(history))

    for i, model in enumerate(history):
        x_data[i], y_data[i] = xy_function(model)

    ax.plot(x_data, y_data, fmt, **kwargs)


def plot_crosshair(ax, x, y, **kwargs):
    lx = ax.axvline(**kwargs)
    lx.set_xdata(x)

    ly = ax.axhline(**kwargs)
    ly.set_ydata(y)


def plot_custom_curve(ax, *args, **kwargs):
    ax.plot(*args, **kwargs)


def plot_bounding_cube(ax, bounding_box, color='w'):
    min_x, max_x, min_y, max_y, min_z, max_z = bounding_box

    xyz_min = np.array([min_x, min_y, min_z])
    xyz_max = np.array([max_x, max_y, max_z])

    max_range = np.array(xyz_max - xyz_min).max()

    center = (xyz_max + xyz_min) / 2

    corners = max_range / 2 * np.mgrid[-1:2:2, -1:2:2, -1:2:2].reshape(3, 8).T + center

    for x, y, z in corners:
        ax.plot([x], [y], [z], color)


def show_load_displacement_curve(model, dof, invert_xaxis=True, block=True):
    dof_type, node_id = dof

    fig, ax = plt.subplots()

    ax.set(xlabel='{} at node {}'.format(dof_type, node_id),
           ylabel=r'Load factor ($\lambda$)',
           title='Load-displacement diagram')
    ax.set_facecolor('white')
    ax.grid()

    plot_load_displacement_curve(ax, model, dof)

    if invert_xaxis:
        ax.invert_xaxis()

    plt.show(block=block)


def show_animation(model, speed=200, block=True):
    if model.status == ModelStatus.eigenvector:
        return show_eigenvector_animation(model, speed, block)
    else:
        return show_history_animation(model, speed, block)


def show_history_animation(model, speed=200, block=True):
    history = model.get_model_history()

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    a = animate_model(fig, ax, history, speed=speed)

    plt.show(block=block)

    return a


def show_eigenvector_animation(model, speed=200, block=True):
    eigenvector = model
    initial_model = model.get_initial_model()

    models = [initial_model, eigenvector]

    bounding_box = get_bounding_box(models)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    def update(step):
        step_model = models[step]

        ax.clear()

        ax.grid()

        plot_bounding_cube(ax, bounding_box)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        plt.title('Eigenvector of Structure \n{}'.format(step_model.name))

        plot_model(ax, step_model, 'gray', True)
        plot_model(ax, step_model, 'red', False)

    a = anim.FuncAnimation(fig, update, frames=2, repeat=True, interval=speed)

    plt.show(block=block)

    return a


def show_deformation_plot(model, step=None, block=True):

    bounding_box = get_bounding_box([model])

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.clear()

    ax.grid()
    plot_bounding_cube(ax, bounding_box)

    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    plot_model(ax, model, 'gray', True)

    if step is None:
        model = model.get_model_history()[step]
    else:
        step = len(model.get_model_history())-1

    plot_model(ax, model, 'red', False)

    ax.set_title('Deformed structure at time step {}\n{}'.format(step, model.name))

    plt.show(block=block)



class Dof:
    def __init__(self, id, value, is_active=True, external_force=0.0):
        self.id = id
        self.reference_value = value
        self.value = value
        self.is_active = is_active
        self.external_force = 0.0

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id
        return self.id == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    @property
    def delta(self):
        return self.value - self.reference_value

    @delta.setter
    def delta(self, value):
        self.value = self.reference_value + value



try:
    from nfem.visualization.interactive import interact # needs to be imported first, because it sets the matplotlib backend
except Exception as _:
    pass

try:
    get_ipython()
    _is_notebook = True
except Exception as _:
    _is_notebook = False

if _is_notebook:
    from nfem.visualization.notebook_animation import show_animation, show_deformation_plot
    from nfem.visualization.notebook_plot import show_load_displacement_curve, Plot2D
else:
    from nfem.visualization.plot import show_load_displacement_curve, show_animation, show_deformation_plot
    from nfem.visualization.plot import Plot2D



"""This module only contains the truss element.

Authors: Thomas Oberbichler, Klaus Sautter
"""

import numpy as np
import numpy.linalg as la


class Truss:
    """FIXME"""

    def __init__(self, id, node_a, node_b, youngs_modulus, area, prestress=0):
        """FIXME"""

        self.id = id
        self.node_a = node_a
        self.node_b = node_b
        self.youngs_modulus = youngs_modulus
        self.area = area
        self.prestress = prestress

    @property
    def dofs(self):
        """FIXME"""
        node_a = self.node_a
        node_b = self.node_b

        return [node_a._dof_x, node_a._dof_y, node_a._dof_z, node_b._dof_x, node_b._dof_y, node_b._dof_z]

    def get_reference_vector(self):
        """FIXME"""

        reference_a = self.node_a.reference_location
        reference_b = self.node_b.reference_location

        return reference_b - reference_a

    def get_actual_vector(self):
        """FIXME"""

        actual_a = self.node_a.location
        actual_b = self.node_b.location

        return actual_b - actual_a

    def get_reference_length(self):
        """FIXME"""

        reference_a = self.node_a.reference_location
        reference_b = self.node_b.reference_location

        return la.norm(reference_b - reference_a)

    def get_actual_length(self):
        """FIXME"""

        actual_a = self.node_a.location
        actual_b = self.node_b.location

        return la.norm(actual_b - actual_a)

    def get_reference_transformMatrix(self):
        """ Transformation matrix for the reference configuration.

        Returns
        -------
        reference_transform : ndarray
            Transformation matrix.
        """
        direction = self.get_reference_vector()
        direction = direction / la.norm(direction)

        reference_transform = np.zeros((2, 6))
        reference_transform[0, :3] = direction
        reference_transform[1, 3:] = direction

        return reference_transform

    def get_actual_transform_matrix(self):
        """ Transformation matrix for the actual configuration.

        Returns
        -------
        actual_transform : ndarray
            Transformation matrix.
        """
        direction = self.get_actual_vector()
        direction = direction / la.norm(direction)

        actual_transform = np.zeros((2, 6))
        actual_transform[0, :3] = direction
        actual_transform[1, 3:] = direction

        return actual_transform

    def calculate_elastic_stiffness_matrix(self):
        """FIXME"""

        e = self.youngs_modulus
        a = self.area
        reference_length = self.get_reference_length()
        reference_transform = self.get_reference_transformMatrix()

        k_e = e * a / reference_length

        return reference_transform.T @ [[k_e, -k_e], [-k_e, k_e]] @ reference_transform

    def calculate_material_stiffness_matrix(self):
        """FIXME"""

        e = self.youngs_modulus
        a = self.area
        actual_length = self.get_actual_length()
        reference_length = self.get_reference_length()
        actual_transform = self.get_actual_transform_matrix()

        k_m = e * a / reference_length * (actual_length / reference_length)**2

        return actual_transform.T @ [[k_m, -k_m], [-k_m, k_m]] @ actual_transform

    def calculate_initial_displacement_stiffness_matrix(self):
        """FIXME"""

        k_m = self.calculate_material_stiffness_matrix()
        k_e = self.calculate_elastic_stiffness_matrix()

        return k_m - k_e

    def calculate_geometric_stiffness_matrix(self, linear=False):
        e = self.youngs_modulus
        a = self.area
        prestress = self.prestress
        reference_length = self.get_reference_length()

        if linear:
            # FIXME this is not a clean solution to solve the LPB issue
            epsilon = self.calculate_linear_strain()
        else:
            epsilon = self.calculate_green_lagrange_strain()

        k_g = e * a / reference_length * epsilon + prestress * a / reference_length

        return np.array([[k_g, 0, 0, -k_g, 0, 0],
                         [0,  k_g, 0, 0, -k_g, 0],
                         [0, 0,  k_g, 0, 0, -k_g],
                         [-k_g, 0, 0,  k_g, 0, 0],
                         [0, -k_g, 0, 0,  k_g, 0],
                         [0, 0, -k_g, 0, 0, k_g]])

    def calculate_stiffness_matrix(self):
        """FIXME"""

        element_k_m = self.calculate_material_stiffness_matrix()
        element_k_g = self.calculate_geometric_stiffness_matrix()

        return element_k_m + element_k_g

    def calculate_green_lagrange_strain(self):
        """FIXME"""

        reference_length = self.get_reference_length()
        actual_length = self.get_actual_length()

        e_gl = (actual_length**2 - reference_length**2) / (2 * reference_length**2)

        return e_gl

    def calculate_linear_strain(self):
        """FIXME"""

        reference_vec = self.get_reference_vector()
        actual_vec = self.get_actual_vector()

        reference_length = self.get_reference_length()

        # project actual on reference
        projected_l = reference_vec @ actual_vec / reference_length

        e_lin = (projected_l - reference_length) / reference_length

        return e_lin

    def calculate_internal_forces(self):
        """FIXME"""

        e_gl = self.calculate_green_lagrange_strain()

        E = self.youngs_modulus
        A = self.area
        prestress = self.prestress

        reference_length = self.get_reference_length()
        actual_length = self.get_actual_length()

        deformation_gradient = actual_length / reference_length

        normal_force = (E * e_gl + prestress) * A * deformation_gradient

        local_internal_forces = [-normal_force, normal_force]

        actual_transform = self.get_actual_transform_matrix()

        global_internal_forces = actual_transform.T @ local_internal_forces

        return global_internal_forces



"""
Interactive
===========

Module to initiate the NFEM-Teching-Tool GUI
"""


import traceback

from nfem.visualization.python_ui import ApplicationWindow, Option, Fore, Style
from nfem.visualization.gui_classes import SideBySide2D3DPlots, AnalysisTab, VisualisationTab, AnimationWindow, set_stiffness_matrix
from nfem.bracketing import bracketing


def interact(model, dof):
    """
    Launch the interactive GUI

    Parameters
    ----------
    model : object
        An object of class Model
    dof : tuple
        a tuple of the node ID and dof type

    Returns
    -------
    model : object
        the model after the GUI is closed

    Examples
    --------
        >>> model = interact(model=model, dof=(42, 'u'))
    """
    window = MainWindow.run(model=model, dof=dof)
    return window.model


class MainWindow(ApplicationWindow):
    """
    Main Application window.

    Parameters
    ----------
    model : Model
        An object of the class model
    dof : tuple
        A tuple of the node id and the dof type
    """

    def __init__(self, model, dof):
        super(MainWindow, self).__init__(
            title=f'NFEM Teaching Tool (Model: {model.name})',
            content=SideBySide2D3DPlots,
            size=(1024, 700))

        self.branches = [model]
        self.dof = dof

        # Check if the given dof is a free dof in the model
        try:
            dof_index = model.dofs.index(dof)
        except ValueError:
            self.WARNING(f'Selected dof {dof} is not part of the dofs: {model.dofs}')
            self.WARNING(f'Running the tool with the first free dof: {model.dofs[0]}')
            dof_index = 0

        self.options = dict()
        # == analysis options
        self.options['tab_idx'] = Option(0)
        self.options['solver_idx'] = Option(0)

        self.options['linear/lambda'] = Option(0)

        self.options['nonlinear/predictor_idx'] = Option(0)
        self.options['nonlinear/predictor/tangential_flag'] = Option(True)
        self.options['nonlinear/predictor/lambda'] = Option(0.0)
        self.options['nonlinear/predictor/delta_lambda'] = Option(0.0)
        self.options['nonlinear/predictor/dof_idx'] = Option(dof_index)
        self.options['nonlinear/predictor/dof_value'] = Option(0.0)
        self.options['nonlinear/predictor/delta-dof'] = Option(0.0)
        self.options['nonlinear/predictor/increment_length'] = Option(0.0)
        self.options['nonlinear/predictor/beta'] = Option(1.0)

        self.options['nonlinear/constraint_idx'] = Option(0)
        self.options['nonlinear/constraint/dof_idx'] = Option(dof_index)

        self.options['nonlinear/newtonraphson/maxiterations'] = Option(100)
        self.options['nonlinear/newtonraphson/tolerance_power'] = Option(-7)

        self.options['nonlinear/solution/det(K)_flag'] = Option(False)
        self.options['nonlinear/solution/eigenproblem_flag'] = Option(False)

        self.options['bracketing/maxiterations'] = Option(100)
        self.options['bracketing/tolerance_power'] = Option(-7)

        self.options['plot/highlight_dof'] = Option(True, self.redraw)
        self.options['plot/dirichlet'] = Option(True, self.redraw)
        self.options['plot/neumann'] = Option(True, self.redraw)
        self.options['plot/symbol_size'] = Option(5, self.redraw)
        self.options['plot/eigenvector_flag'] = Option(False, self.redraw)
        self.options['plot/dof_idx'] = Option(dof_index, self.redraw)
        self.options['plot/load_disp_curve_flag'] = Option(True, self.redraw)
        self.options['plot/load_disp_curve_iter_flag'] = Option(False, self.redraw)
        self.options['plot/det(K)_flag'] = Option(False, self.redraw)
        self.options['plot/eigenvalue_flag'] = Option(False, self.redraw)

        self.options['stiffness/system_idx'] = Option(
            value=0,
            action=lambda: set_stiffness_matrix(
                model=self.model,
                debugger=self.DEBUG_blue,
                **self.options
                )
            )
        self.options['stiffness/component_idx'] = Option(
            value=0,
            action=lambda: set_stiffness_matrix(
                model=self.model,
                debugger=self.DEBUG_blue,
                **self.options
                )
            )
        self.options['stiffness/matrix'] = Option([])

        self.redraw()

    @property
    def option_values(self):
        """ dict of self.options values instead of objects """
        return {option[0]: option[1].value for option in self.options.items()}

    @property
    def solvers(self):
        return ['nonlinear', 'LPB', 'linear', 'bracketing']

    @property
    def predictors(self):
        return [
            'lambda',
            'delta-lambda',
            'dof',
            'delta-dof',
            'arc-length',
            'increment',
            'arclength_eigenvector'
        ]

    @property
    def constraints(self):
        return ['load-control', 'displacement-control', 'arc-length-control']

    @property
    def model(self):
        return self.branches[-1]

    @model.setter
    def model(self, value):
        self.branches[-1] = value

    def DEBUG(self, message=None):
        """ print to console """
        if message:
            print(message)

    def DEBUG_red(self, message=None):
        """ print to console in red color"""
        if message:
            print(Fore.RED + message + Style.RESET_ALL)

    def DEBUG_blue(self, message=None):
        """ print to console in blue color"""
        if message:
            print(Fore.CYAN + message + Style.RESET_ALL)

    def WARNING(self, message=None):
        """ print WARNING to console"""
        if message:
            print(Fore.YELLOW + "WARNING: " + message + Style.RESET_ALL)

    def _build_sidebar(self, builder):
        builder.add_tabs(
            items=[
                ('Analysis', AnalysisTab),
                ('Visualisation', VisualisationTab)],
            option=self.options['tab_idx'])
        builder.add_stretch()
        builder.add_button(
            label='Solve',
            action=self.solve_click)
        builder.add_button(
            label='Go back',
            action=self.go_back_click)
        builder.add_button(
            label='Reset path',
            action=self.reset_path_click)
        builder.add_button(
            label='New path',
            action=self.new_path_click)
        builder.add_button(
            label='Reset all',
            action=self.reset_all_click)

    def redraw(self):
        self.content.redraw()

    def show_animation_click(self, builder):
        builder.show_dialog(AnimationWindow, size=(700, 700))

    def solve_click(self):
        try:
            solver = self.solvers[self.options['solver_idx'].value]

            if solver == 'linear':
                model = self.solve_linear(
                    model=self.model.get_duplicate(),
                    lam=self.options['linear/lambda'].value)

            elif solver == 'LPB':
                model = self.solve_LPB(
                    model=self.model.get_duplicate(),
                    lam=self.options['linear/lambda'].value)

            elif solver == 'nonlinear':
                model = self.solve_nonlinear(
                    model=self.model.get_duplicate(),
                    options=self.option_values)

            elif solver == 'bracketing':
                model = self.solve_bracketing(
                    model=self.model,
                    options=self.option_values)

            else:
                raise Exception(f'Unknown solver {solver}')

        except Exception as e:
            traceback.print_exc()
            self.show_error_dialog(str(e))
            return

        self.model = model
        self.options['nonlinear/predictor/increment_length'].change(model.get_increment_norm())
        self.redraw()

    def solve_linear(self, model, lam):
        model.load_factor = lam
        model.perform_linear_solution_step()
        return model

    def solve_LPB(self, model, lam):
        if model.get_previous_model().get_previous_model() is not None:
            raise RuntimeError('LPB can only be done on the initial model')
        model.load_factor = lam
        model.perform_linear_solution_step()
        model.solve_linear_eigenvalues()
        return model

    def solve_nonlinear(self, model, options):
        predictor = self.predictors[options['nonlinear/predictor_idx']]
        tangential_flag = options['nonlinear/predictor/tangential_flag']

        if predictor == 'lambda':
            value = options['nonlinear/predictor/lambda']
            if tangential_flag:
                model.predict_tangential(strategy=predictor, value=value)
            else:
                model.load_factor = value

        elif predictor == 'delta-lambda':
            value = options['nonlinear/predictor/delta_lambda']
            if tangential_flag:
                model.predict_tangential(strategy=predictor, value=value)
            else:
                model.load_factor += value

        elif predictor == 'dof':
            dof = model.dofs[options['nonlinear/predictor/dof_idx']]
            dof_value = options['nonlinear/predictor/dof_value']
            if tangential_flag:
                model.predict_tangential(strategy=predictor, dof=dof, value=dof_value)
            else:
                model[dof].delta = dof_value

        elif predictor == 'delta-dof':
            dof = model.dofs[options['nonlinear/predictor/dof_idx']]
            dof_value_increment = options['nonlinear/predictor/delta-dof']
            if tangential_flag:
                model.predict_tangential(strategy=predictor, dof=dof, value=dof_value_increment)
            else:
                model[dof].delta += dof_value_increment

        elif predictor == 'arc-length':
            arclength = options['nonlinear/predictor/increment_length']
            model.predict_tangential(strategy=predictor, value=arclength)

        elif predictor == 'increment':
            increment_length = options['nonlinear/predictor/increment_length']
            model.predict_with_last_increment(value=increment_length)

        elif predictor == 'arclength_eigenvector':
            arclength = options['nonlinear/predictor/increment_length']
            model.predict_tangential(strategy='arc-length', value=arclength)
            beta = options['nonlinear/predictor/beta']
            model.combine_prediction_with_eigenvector(beta)

        else:
            raise Exception('Unknown predictor {}'.format(predictor))

        constraint = self.constraints[options['nonlinear/constraint_idx']]
        constraint_dof = model.dofs[options['nonlinear/constraint/dof_idx']]
        tolerance_power = options['nonlinear/newtonraphson/tolerance_power']
        max_iterations = options['nonlinear/newtonraphson/maxiterations']
        determinant = options['nonlinear/solution/det(K)_flag']
        eigenproblem = options['nonlinear/solution/eigenproblem_flag']

        model.perform_non_linear_solution_step(
            strategy=constraint,
            tolerance=10**tolerance_power,
            dof=constraint_dof,
            max_iterations=max_iterations,
            solve_det_k=determinant,
            solve_attendant_eigenvalue=eigenproblem)

        return model

    def solve_bracketing(self, model, options):
        backup_model = model.get_duplicate(branch=True)

        try:
            model = bracketing(
                model,
                tol=10**options['bracketing/tolerance_power'],
                max_steps=options['bracketing/maxiterations'],
                raise_error=True,
                tolerance=10**options['nonlinear/newtonraphson/tolerance_power'],
                max_iterations=options['nonlinear/newtonraphson/maxiterations'],
                solve_det_k=options['nonlinear/solution/det(K)_flag'],
                solve_attendant_eigenvalue=options['nonlinear/solution/eigenproblem_flag']
            )
        except Exception as e:
            traceback.print_exc()
            self.show_error_dialog(str(e))
            return backup_model

        return model

    def go_back_click(self):
        if self.model.get_previous_model() is None:
            return
        self.model = self.model.get_previous_model()
        self.redraw()
        self.DEBUG_red('Went back one step!')

    def reset_path_click(self):
        if self.model.get_previous_model() is None:
            return
        self.model = self.model.get_initial_model()
        self.redraw()
        self.DEBUG_red('Path has been reset!')

    def new_path_click(self):
        new_model = self.model.get_duplicate(branch=True)
        self.branches.append(new_model)
        self.redraw()
        self.DEBUG_blue('Created a new path!')

    def reset_all_click(self):
        model = self.model.get_initial_model()
        self.branches = [model]
        self.redraw()
        self.DEBUG_red('Reset all has been clicked!')



import plotly.graph_objects as go
import numpy as np


class Plot2D:
    def __init__(self, x_label='Displacement', y_label=r'Load factor (λ)', title='Load-displacement diagram'):
        self.title = title
        self.x_label = x_label
        self.y_label = y_label
        self.invert_xaxis = True
        self.data = []

    def add_load_displacement_curve(self, model, dof, label=None, show_iterations=False):
        _add_load_displacement_curve(self.data, model, dof, label)

        if show_iterations:
            _plot_load_displacement_iterations(self.data, model, dof, label)

    def add_det_k_curve(self, model, dof, label=None):
        _plot_det_k_curve(self.data, model, dof, label)

    def add_history_curve(self, model, fun, label, show_iterations=False):
        _plot_history_curve(self.data, model, fun, label, not show_iterations)

    def add_custom_curve(self, x, y, label, linewidth=1.0):
        self.data.append(go.Scatter(
            name=label,
            mode='lines+markers',
            x=x,
            y=y,
            line=dict(
                width=linewidth,
            ),
        ))

    def show(self, height=500):
        fig = go.Figure(
            data=self.data,
            layout=go.Layout(
                title=self.title,
                xaxis=dict(
                    autorange='reversed',
                    title_text=self.x_label,
                ),
                yaxis=dict(
                    title_text=self.y_label,
                ),
            ),
        )
        fig.show()


def show_load_displacement_curve(model, dof):
    dof_type, node_id = dof

    plot = Plot2D()
    model.add_load_displacement_curve(model, dof)

    plot.show()


def _add_load_displacement_curve(data, model, dof, label):
    history = model.get_model_history()

    x = np.zeros(len(history))
    y = np.zeros(len(history))

    for i, model in enumerate(history):
        x[i] = model[dof].delta
        y[i] = model.load_factor

    node_id, dof_type = dof

    if label is None:
        label = f'λ : {dof_type} at node {node_id}'

    data.append(go.Scatter(
        name=label,
        mode='lines+markers',
        x=x,
        y=y,
    ))


def _plot_load_displacement_iterations(data, model, dof, label):
    history = model.get_model_history(skip_iterations=False)

    x = np.zeros(len(history))
    y = np.zeros(len(history))

    node_id, dof_type = dof

    for i, model in enumerate(history):
        x[i] = model[dof].delta
        y[i] = model.load_factor

    if label is None:
        label = f'λ : {dof_type} at node {node_id} (iter)'
    else:
        label += ' (iter)'

    data.append(go.Scatter(
        name=label,
        mode='lines+markers',
        x=x,
        y=y,
        line=dict(
            width=0.75,
        ),
        marker=dict(
            size=2.0,
        ),
    ))


def _plot_det_k_curve(data, model, dof, label):
    history = model.get_model_history()

    x = np.zeros(len(history))
    y = np.zeros(len(history))

    node_id, dof_type = dof

    for i, model in enumerate(history):
        x[i] = model[dof].delta
        y[i] = model.det_k

    if label is None:
        label = 'det(K) : {} at node {}'.format(dof_type, node_id)

    data.append(go.Scatter(
        name=label,
        mode='lines+markers',
        x=x,
        y=y,
    ))


def _plot_history_curve(data, model, fun, label, skip_iterations):
    history = model.get_model_history(skip_iterations)

    x = np.zeros(len(history))
    y = np.zeros(len(history))

    for i, model in enumerate(history):
        x[i], y[i] = fun(model)

    data.append(go.Scatter(
        name=label,
        mode='lines+markers',
        x=x,
        y=y,
    ))



from setuptools import setup

import os
import re


def get_version(path):
    with open(os.path.join(os.path.dirname(__file__), path)) as file:
        data = file.read()
    regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(regex, data, re.M)
    if version_match is None:
        raise RuntimeError("Unable to find version string.")
    return version_match.group(1)


setup(
    name='nfem',
    version=get_version(os.path.join('nfem', '__init__.py')),
    description='NFEM Teaching Tools',
    url='https://github.com/StatikTUM/nfem',
    author='Thomas Oberbichler, Armin Geiser, Klaus Sautter, Aditya Ghantasala, Mahmoud Zidan',
    author_email='',
    license='',
    packages=['nfem', 'nfem.visualization'],
    python_requires='>3.6',
    install_requires=[
        'ipython',
        'matplotlib',
        'numpy',
        'plotly',
        'pyqt5',
        'scipy',
    ],
    zip_safe=False,
)
